@@ -1,10 +1,10 @@
 import threading
 import weakref
-from typing import Generic, TypeVar
+from typing import Generic, Optional, TypeVar
 
 from streamlit.report_session import ReportSession
 
-from .session_info import get_this_session_info
+from .hash import Hash, calc_hash
 
 StateValueT = TypeVar("StateValueT")
 
@@ -16,6 +16,7 @@ class ValueNotSetError(Exception):
 class ServerStateItem(Generic[StateValueT]):
     _is_set: bool
     _value: StateValueT
+    _value_hash: Optional[Hash]
     _value_lock: threading.RLock
 
     _bound_sessions: "weakref.WeakSet[ReportSession]"
@@ -23,34 +24,35 @@ class ServerStateItem(Generic[StateValueT]):
 
     def __init__(self) -> None:
         self._is_set = False
+        self._value_hash = None
         self._value_lock = threading.RLock()
 
         self._bound_sessions = weakref.WeakSet()
         self._bound_sessions_lock = threading.Lock()
 
-    def setup_for_this_session(self) -> None:
-        this_session_info = get_this_session_info()
-        if this_session_info is None:
-            raise RuntimeError(
-                "Oh noes. Couldn't get your Streamlit Session object. "
-                "Are you doing something fancy with threads?"
-            )
-        this_session = this_session_info.session
+    def bind_session(self, session: ReportSession) -> None:
         with self._bound_sessions_lock:
-            if this_session not in self._bound_sessions:
-                self._bound_sessions.add(this_session)
+            if session not in self._bound_sessions:
+                self._bound_sessions.add(session)
 
     def _rerun_bound_sessions(self) -> None:
         with self._bound_sessions_lock:
             for session in self._bound_sessions:
                 session.request_rerun(client_state=None)  # HACK: XD
 
+    def _on_set(self):
+        new_value_hash = calc_hash(self._value)
+        if self._value_hash is None or self._value_hash != new_value_hash:
+            self._rerun_bound_sessions()
+
+        self._value_hash = new_value_hash
+
     def set_value(self, value: StateValueT) -> None:
         with self._value_lock:
             self._is_set = True
             self._value = value
 
-        self._rerun_bound_sessions()
+        self._on_set()
 
     def get_value(self) -> StateValueT:
         with self._value_lock:
